#!/usr/bin/python
#####   Tile(set) editor v5.3   ##############################################
### Riley Ellis, 26 August 2009
#	Based upon 'spriteedit' (4.2), started [28 July 2009]
#		and 'tileset' (3.0), started [20 August 2009]
#		and 'Editor' (2.0), started [24 August 2009]
#	Completely re-written to accomodate new ideas
#	  - but borrowed a lot of old code
#
## To-do:
#	x Some kind of menus
#		Edit menu: Undo, Redo, Revert, Clear, Invert, Fill
#	x Invert (no caller yet) [02:57 PM, 10-24-09] 
#	x Set all black (no caller yet) [03:05 PM, 10-24-09] 
#	x Reflect/rotate [08:26 PM, 12-28-09] 
#	- Have single sprite, or full tileset mode for sEditor
#	x Undo/redo [07:12 PM, 12-28-09]
#	- Modify to allow editing pictures
#	- Side palette with editing options
#	- For resizing, have it truncate current sprite for shrinking,
#		- and add blank space for expanding
#		- or prompt user for what they want to do?
#	- Rework initializing code so get_size is called after creation
#		of master window but before creation of edit_field
#
## Ideas:
#	x Split the GUI and actual functionality (all these re-works are ridiculus!) [07:12 PM, 12-28-09] 
#	x Rename "undone" to "redos" (less confusing)
#	@ Allow for tiles to not all be the same size? (screw it) [02:50 PM, 10-21-09] 
#	x Don't require user to quit and restart to change size of tiles
#	- Grayscale eventually?
#	- Non-byte-wide tiles?
#	- Option to compare old tile to new?
#	- View different data formats (binary/hex strings; ASM format, ...)
#	- Export to .8xi pictures?
#	- Deal with bad tile (data) inputs
#	- Larger "brush size" (one pixel at a time on a 24x24 tile sucks)
#	- Allow for changing of pixel box size?
#
## Notes:
#	- Intended to be self-contained, but also able to be called from
#		other programs (e.g. Tilemap editor)
#	- 'Quit' and 'Okay' will function differently for tiles and tilesets
#	- For now, the only option for export is to a hex string
#
## Info:
#
##############################################################################


## x Working on Tile class to make some things independent of sEdit [03:07 PM, 10-21-09] 
#	x has enough features for use with sEditor [04:22 PM, 10-21-09] 
## x Need to implement tile initialization [09:39 AM, 10-22-09] 
#
## x Implement edit methods
#	x All but 'update' working (well, it's crap that works) [09:54 AM, 10-22-09] 
#	x ... it's not too crappy; nevermind [02:38 PM, 10-22-09] 
#
## @ Implement UI methods
#	@ Front-end stuff moved to Editor.py (FSE,2) [03:30 PM, 10-22-09] 
#
## x Currently working on the front-end stuff [03:31 PM, 10-22-09] 
#
## x Implementing undo/redo [03:42 PM, 10-24-09] 
#	x My current undo/redo logic is wrong [06:03 PM, 10-27-09] 
#		^^ Fixed [09:17 AM, 11-03-09]
#	x Need to implement adding actions to the undos stack
#	x Finished (I think) [09:57 AM, 11-03-09] 
#
## x Adding menus and cleaning up the interface [07:39 PM, 12-28-09] 
#	x Menus [09:02 PM, 12-28-09] 
#	x Shortcuts and such [03:43 PM, 12-30-09] 
#


## Planned format of window:
#########################################################
#   Name [entry]   Number [entry]			#
# T-------------------------------T  <Clear>		#
# l				  l  <Revert>		#
# l				  l			#
# l	[edit sprite field]	  l  [current sprite]	#
# l				  l			#
# l				  l  <OK>		#
# L-------------------------------4  <Exit>		#
#			    (x, y)			#
#-------------------------------------------------------#
# [tile palette]					#
#########################################################

## Needed data/formats for the sprite data:
#	- Name, number
#	- Width, height
#	- XBM format
#	- String format (for storage)
#	- Array of binary digits (char strings)
#	- Asm format
## Needed methods for sprite data:
#	- Set/reset pixels
#	- Convert  bArray->String, bArray->XBM, String->bArray, ->Asm
#	- Scale


## Things that can be undone/redone:
#	- Setting/resetting pixels
#	- Clearing
#	- Reverting
#	- Inverting
#	- Filling
## Setting/resetting pixels will be grouped until button released
#	i.e. dragging counts as one action
#
## I'll just store whole sprite since dealing with pixel-by-pixel or
#	bunch-o-pixel would be cumbersome [05:24 PM, 10-27-09] 
#
## Action (if different):
#	Push old onto undos
#	Truncate to 20
#	Update sprite/screen (with new)
#	Clear undone (no more redos)
## Undo (if any):
#	Push current onto undone
#	Pop from undos
#	 and update sprite/screen
## Redo (if any):
#	Push current onto undos
#	Pop from undone
#	 and update sprite/screen

##############################################################################

from Tkinter import (Tk, Frame, Label, Entry, Button, Canvas, 
	BitmapImage, StringVar, IntVar,
	Toplevel)
from tkMessageBox import showinfo
from sEdit_misc import (bitmapstring, colors, SEP, BOX, filler, splice,
	nbin, s_sum, UR_MAX, BINDINGS, C_STATES)
from sEditUI import sEditUI
from sEdit_input import get_size


##############################################################################
###  sEditor  ################################################################
class sEditor(sEditUI):		# Base class for tile and sprite editors
	def __init__(self, width, height, master = None, **args):
		# width/height = tile size; width must be multiple of 8
		#  ^ user-given width/height is useless because get_size is always called
		# master = parent window (must accomodate being called from external program)
		#	^ thinking of not using master
		# args = catch-all, placeholder, and reminder for other arguments
		self.width, self.height = width, height

		# Note: cur_tile is initialized in the calling Editor
		sEditUI.__init__(self, width, height)
		self.set_tile(self.cur_tile)	# This also calls update

#		self.debug1()
#		self.palette_init()


	def set_tile(self, tile, name = None, num = None):  # sets new cur_tile, and mods def_tile, name and num
		# I use a tile instead of data to ease integrating this with map editor
		self.ur_init()	# new tile -> new undo/redo
		self._set_states()	# update the undo/redo buttons

		if name != None:	# if user gave name
			self.cur_name.set(name)
		if num != None:		# if user gave number
			self.cur_num.set(num)
		self.cur_tile = tile
		self.def_tile = tile.__str__()	# initial version of tile; used by "revert" (DEFault)

		# v disable rotations if tile isn't square
		rotate = C_STATES[self.width == self.height]	# normal or disabled
		self.editmenu.entryconfig(12, state = rotate)
		self.editmenu.entryconfig(13, state = rotate)

		self.update()

	def resize(self):
		new_size = get_size(self.master, self.width, self.height)
		if new_size == (self.width, self.height):
			return	# don't do anything if size didn't change

		self.width, self.height = new_size
		map(lambda i: self.edit_field.unbind(i), BINDINGS)	# get rid of bindings on edit_field

		self.edit_field.destroy()
		self.editField_init(*new_size)	# create new edit_field
		self.set_tile(Tile(*new_size))	# set blank tile


	def get(self):	# string of current tile; used often
		return self.cur_tile.__str__()
	def set(self, data):	# sets current sprite
		self.cur_tile.fromstring(data)


	######################################################################
	###  User commands and such  #########################################
	def control_binds(self, event = 0):	# <Control-*> calls this
		key = event.keysym.lower()
		if key == 'n':		self.new(event)
		elif key == 'o':	self.open(event)
		elif key == 's':	self.save(event)
		elif key == 'e':	self.export(event)
		elif key == 'q':	self.quit(event)
		elif key == 'z':	self.undo(event)


	def run(self):	# only for development
		self.master.mainloop()

	def okay(self, event = 0):	# Different for Tiles and Tilesets
		self.def_tile = self.cur_tile.__str__()	# sets current tile to default (for revert)
		pass	# Maybe use as save button? or set current to default tile?
	def quit(self, event = 0):	# Different for Tiles and Tilesets?
		self.master.quit()

	def new(self, event = 0):
		self.resize()	# prompts for new size and clear current one

	def open(self, event = 0):
		print "Why the hell do you want to open another one?"
	def save(self, event = 0):
		print "You really like it that much?"

	def export(self, event = 0):	# I'm only giving this one option for now
		print ''.join(map(lambda i: hex(ord(i)+0x100)[-2:].upper(), self.get()))

	def help(self, event = 0):
		showinfo(parent = self.master, title = "Help", message = """
			Things that aren't too obvious:
			  Left-click:  Set pixel
			  Right-click: Clear pixel
			  'Okay': Sets current tile as default (for revert)""")
	def about(self, event = 0):
		print "This program does stuff."

	######################################################################


	######################################################################
	###  Sprite-altering methods  ########################################
	def clear(self, event = 0):	# resets all bits
		self.cur_tile.fill(0)
		self.modify()	# update the undo/redo stack, plus display
	def fill(self, event = 0):	# sets all bits
		self.cur_tile.fill(1)
		self.modify()

	def revert(self, event = 0):	# reverts cur_tile to it's initial state
		self.cur_tile.fromstring(self.def_tile)
		self.modify()

	def invert(self, event = 0):	# makes dark pixels light, and light pixels dark
		self.cur_tile.invert()
		self.modify()

	def vflip(self, event = 0):	# flip about vertical axis
		self.cur_tile.flip_vert()
		self.modify()
	def hflip(self, event = 0):	# flip about horizontal axis
		self.cur_tile.flip_horiz()
		self.modify()

	def ccRotate(self, event = 0):
		self.cur_tile.rotate(1)
		self.modify()
	def clRotate(self, event = 0):
		self.cur_tile.rotate(-1)
		self.modify()


	def setp(self, event = 0):	# sets pixel (duh)
		self.modp(1, event)
	def clearp(self, event = 0):	# clears pixel (double duh)
		self.modp(0, event)

	def modp(self, value, event = 0):	# modifies a pixel
		# event.type == 4 (click), 6 (motion)
		item = self.edit_field.find_closest(event.x, event.y)[0]
		(row, col) = divmod(item - 2, self.width)

		if (item != 1) and (item <= 1 + self.width*self.height):	# the background/space between "pixels"
			self.cur_tile[row, col] = value
			self.edit_field.itemconfigure(item, fill = colors[value])
			self.show()
		self.curdisp(event)
	######################################################################
	###  Undo/redo methods  ##############################################
	def ur_init(self, event = 0):	# initialize undo/redo
		self.ur_stack = [self.get()]
		self.ur_ptr = 0

	def undo(self, event = 0):
		if self.ur_ptr:
			self.ur_do(-1)
	def redo(self, event = 0):
		if self.ur_redos:
			self.ur_do(1)
	def ur_do(self, dir, event = 0):	# does the actual changes due to undo/redo
		self.ur_ptr += dir
		self.set(self.ur_stack[self.ur_ptr])
		self._set_states()
		self.update()
	def modify(self, event = 0):	# called to update the undo/redo stack
		if self.ur_stack[self.ur_ptr] != self.get():
			if self.ur_ptr == UR_MAX:
				self.ur_ptr -= 1
				self.ur_stack.pop(0)
			self.ur_ptr += 1
			self.ur_stack = self.ur_stack[:self.ur_ptr] + [self.get()]

			self._set_states()
			self.update()
	######################################################################

		
##############################################################################


##############################################################################
## Needed data/formats for the sprite data:
#	@ Name, number (external to Tile class)
#	x Width, height
#	x XBM format
#	x String format (for storage)
#	- Array of binary digits (char strings)
#	? Asm format (can't decide whether external or internal)
## Needed methods for sprite data:
#	x Set/reset pixels
#	x Convert  bArray->String, bArray->XBM, String->bArray, (->Asm)
#	x Scale
##############################################################################
###  Tile  ###################################################################
class Tile:
	def __init__(self, width, height, data = None):
		# width, height = size in pixels
		# data = initial data; in char string format
		self.width, self.height = width, height
		if not data:	# make it blank
			data = '\x00' * (height * width/8)

		self.fromstring(data)	# sets self.sprite

	def __setitem__(self, (row, bit), value):	# sets (row, bit) to value
		t_row = self.sprite[row]
		self.sprite[row] = t_row[:bit] + str(value) + t_row[bit+1:]
	def __getitem__(self, (row, bit)):
		return int(self.sprite[row][bit])

#	def getcol(self, col):	# gets the column as a string
#		return ''.join(map(lambda i: i[col], self.sprite))

	def bytes(self):
		for row in self.sprite:
			for byte in splice(row, 8):
				yield byte


	def fill(self, state):	# sets all bits to 'state'
		filler = '1' if state else '0'
		self.sprite = [filler * self.width] * self.height

	def invert(self):	# inverts all bits
		sprite = [ chr(ord(byte) ^ 0xFF) for byte in self.__str__() ]
		self.fromstring(sprite)

	def flip_horiz(self):
		self.sprite = map(lambda i: i[::-1], self.sprite)
	def flip_vert(self):
		self.sprite = self.sprite[::-1]

	def rotate(self, dir):	# turns 90 degrees clockwise (dir == -1 for clockwise, 1 for counter-clockwise)
		self.sprite = map(lambda col: ''.join(map(lambda i: i[col], 
			self.sprite[::dir])), range(self.width)[::-dir])
		# ^ for clockwise, row 'n' is the reverse of column 'n'
		#	for counter-clockwise, row 'n' is column 'width - n'
		


	def fromstring(self, data):	# sets self.sprite using data (size assumed correct)
		self.sprite = splice(s_sum(map(nbin, data)), self.width)
	def __str__(self):	# returns character string representing data
		return map(lambda i: chr(int(i, 2)), self.bytes())

	def tobitmap(self, factor = 1):	# creates xbm bitmap; scaled by factor
		return tobitmap(scale(self.sprite, factor))


def bytes(sprite):	# generator that returns every 8 bits of sprite
	for row in sprite:
		for byte in splice(row, 8):
			yield byte

def tobitmap(sprite):	# creates xbm bitmap
	s_bytes = [ hex(int(i[::-1], 2)) for i in bytes(sprite) ]
	return bitmapstring.format(len(sprite[0]), len(sprite), ', '.join(s_bytes))

def scale(sprite, factor):	# need to find better way
	t_scale = lambda row: s_sum([ i*factor for i in row ])
	t_map = map(t_scale, sprite)
	return s_sum([ [i]*factor for i in t_map ])	# returns in sprite format

##############################################################################





