#!/usr/bin/python
#####   UI code for sEdit.py   ###############################################
##	 (to separate the (mostly) static GUI code from highly dynamic
#	 functionality code)
### Riley Ellis, 28 December 2009
#
## To-do:
#
## Ideas:
#
## Notes:
#
##############################################################################
from Tkinter import (Frame, Tk, StringVar, IntVar, Label, Entry, Canvas, 
	Button, BitmapImage, Toplevel, Menu, PhotoImage)
from sEdit_misc import (SEP, BOX, C_STATES, filler, 
	g_coords, colors, menu_font, focusin)



class sEditUI:
	def __init__(self, width, height):
		self.master = Tk()
		self.master.title("Editor")

		self.master.event_add("<<B1>>", "<Button-1>", "<B1-Motion>")
		self.master.event_add("<<B3>>", "<Button-3>", "<B3-Motion>")

		container = Frame(self.master, relief = 'ridge', border = 3)	# master widget (parent to all)
		container.pack(side = 'top')	# Let user specify where/how to pack?

		## sprite_name sprite_number
		self.cur_name = StringVar(value = '')	# name of current sprite
		self.cur_num = IntVar(value = 0)	# number of current sprite
		filler(container, height = 5, side = 'top')	# space on top

		dataFrame = Frame(container, relief = 'solid', border = 0, padx = 1, pady = 1)
		dataFrame.pack(side = 'top')	# Frame for holding the name and number stuff

		Label(dataFrame, text = "Name").pack(side = 'left')
		name_entry = Entry(dataFrame, textvariable = self.cur_name, width = 7)
		name_entry.bind("<FocusIn>", focusin)
		name_entry.pack(side = 'left')

		filler(dataFrame, width = 10, side = 'left')	# space between Name and Number fields

		Label(dataFrame, text = "Number").pack(side = 'left')
		num_entry = Entry(dataFrame, textvariable = self.cur_num, width = 3)
		num_entry.bind("<FocusIn>", focusin)
		num_entry.pack(side = 'left')
		## \sprite_name sprite_number


		#### edit_field and cursor_coords
		spriteFrame = Frame(container, pady = 5)	# will contain the edit field and coordinates
		spriteFrame.pack(side = 'left')
		self.spriteFrame = spriteFrame

		self.edit_frame = Frame(spriteFrame)	# because of the way I get cursor coords, edit_field
		self.edit_frame.pack(side = 'top')	#	...has to be recreated for resizing of sprite
		self.editField_init(self.width, self.height)


		self.coords = StringVar(value = "(0, 0)")
		curFrame = Frame(spriteFrame)	# holds cursor coord display
		curFrame.pack(side = 'top', fill = 'x')
		Label(curFrame, textvariable = self.coords, justify = 'left').pack(side = 'right')
		## \edit_field cursor_coords


		#### Buttons sprite_display
		dispFrame = Frame(container, padx = 5, pady = 5)
		dispFrame.pack(side = 'left', fill = 'y')
		Button(dispFrame, text = "Clear", command = self.clear).pack(side = 'top', fill = 'x')
		Button(dispFrame, text = "Revert", command = self.revert).pack(side = 'top', fill = 'x')

		filler(dispFrame, height = 8, side = 'top', fill = 'x')	# space between buttons and image
		self.cur_pic = BitmapImage(data = self.cur_tile.tobitmap(3))

		self.spriteLabel = Label(dispFrame, image = self.cur_pic, relief = 'groove', border = 2)
		self.spriteLabel.pack_propagate(0)	# keep from changing size
		self.spriteLabel.pack(side = 'top')
		filler(dispFrame, height = 8, side = 'top', fill = 'x')	# space between image and more buttons

		Button(dispFrame, text = "Okay", command = self.okay).pack(side = 'top', fill = 'x')
		Button(dispFrame, text = "Exit", command = self.quit).pack(side = 'top', fill = 'x')
		## \buttons sprite_display


		#### Bindings and such
		# Note: edit_field bindings are done in editField_init

		# Upper and lower case count as different events
		self.bind("<Control-Key>", self.control_binds)
		self.bind("<Shift-Control-Z>", self.redo)
		self.bind("<Shift-Control-z>", self.redo)
		self.bind("<F1>", self.help)

		self.master.focus_set()
		## \Bindings

		self.menu_init()


	def editField_init(self, width, height):
		self.edit_field = Canvas(self.edit_frame, width = 7 + width*BOX, height = 7 + height*BOX,
			relief = 'solid', border = 2)	# where the pixels will be displayed
		self.edit_field.pack_propagate(0)	# make sure the size doesn't change
		self.edit_field.pack(side = 'top', padx = 3)
		self.edit_field.create_rectangle(0, 0, self.edit_field.cget('width'),
			self.edit_field.cget('height'), width = 0, tag="back")	# background
		for row in range(height):
			for col in range(width):
				self.edit_field.create_rectangle(*g_coords(row, col), 
					outline = 'gray', width = 1)
#					tag = "{0}.{1}".format(row, col))

		for index in range(4, width, 4):	# vertical separator lines
			self.edit_field.create_line(6+BOX*index, 6, 6+BOX*index, 6+BOX*height, fill = SEP)
		for index in range(4, height, 4):	# horizontal separator lines
			self.edit_field.create_line(6, 6+BOX*index, 6+BOX*width, 6+BOX*index, fill = SEP)
		self.edit_field.create_rectangle(6, 6, 6+BOX*width, 6+BOX*height, outline = SEP)

		self.edit_field.bind("<<B1>>", self.setp)
		self.edit_field.bind("<<B3>>", self.clearp)
		self.edit_field.bind("<Motion>", self.curdisp)
		self.edit_field.bind("<ButtonRelease>", self.modify)	# update the undo/redo stack



	def menu_init(self):	# creates the menu for the Editor
		# Don't forget that Tilesets will need extra functions/menus
		menu = Menu(self.master, font = menu_font)

		filemenu = Menu(menu, tearoff = 0, font = menu_font)
		filemenu.add_command(label = "New", command = self.new, accelerator = "Ctrl+N", underline = 0)
		filemenu.add_command(label = "Open...", command = self.open, accelerator = "Ctrl+O", underline = 0)
		filemenu.add_separator()
		filemenu.add_command(label = "Save", command = self.save, accelerator = "Ctrl+S", underline = 0)
		filemenu.add_command(label = "Save As...", command = None)
		filemenu.add_separator()
		filemenu.add_command(label = "Export...", command = self.export, accelerator = "Ctrl+E", underline = 0)
		filemenu.add_separator()
		filemenu.add_command(label = "Quit", command = self.quit, accelerator = "Ctrl+Q", underline = 0)
		menu.add_cascade(label = "File", menu = filemenu)

		self.editmenu = editmenu = Menu(menu, tearoff = 0, font = menu_font)
		editmenu.add_command(label = "Undo", command = self.undo, accelerator = "Ctrl+Z")
		editmenu.add_command(label = "Redo", command = self.redo)
		editmenu.add_separator()
		editmenu.add_command(label = "Resize", command = self.resize)
		editmenu.add_separator()
		editmenu.add_command(label = "Clear", command = self.clear)
		editmenu.add_command(label = "Fill", command = self.fill)
		editmenu.add_command(label = "Invert", command = self.invert)
		editmenu.add_command(label = "Revert", command = self.revert)
		editmenu.add_separator()
		editmenu.add_command(label = "Flip Vertically", command = self.vflip)
		editmenu.add_command(label = "Flip Horizontally", command = self.hflip)
		editmenu.add_command(label = "Rotate Clockwise", command = self.ccRotate)
		editmenu.add_command(label = "Rotate Counter-Clockwise", command = self.clRotate)
		menu.add_cascade(label = "Edit", menu = editmenu)

		helpmenu = Menu(menu, tearoff = 0, font = menu_font)
		helpmenu.add_command(label = "Help", command = self.help, accelerator = "F1")
		helpmenu.add_separator()
		helpmenu.add_command(label = "About", command = self.about)
		menu.add_cascade(label = "Help", menu = helpmenu)

		# these won't always be here; eventually they'll be included
		#	in the editing sidebar
		menu.add_command(bitmap = "@./Subfiles/ccrotate.xbm", command = self.ccRotate)
		menu.add_command(bitmap = "@./Subfiles/clrotate.xbm", command = self.clRotate)
		menu.add_command(bitmap = "@./Subfiles/hflip.xbm", command = self.hflip)
		menu.add_command(bitmap = "@./Subfiles/vflip.xbm", command = self.vflip)

		self.master.config(menu = menu)

	def palette_init(self):	# creates the side bar with editing options
		self.palette = palette = Toplevel(self.master)
		palette.title("Tools")

		Button(palette, bitmap = "@./Subfiles/ccrotate.xbm", 
			command = self.ccRotate).grid(row = 0, column = 0)
		Button(palette, bitmap = "@./Subfiles/clrotate.xbm", 
			command = self.clRotate).grid(row = 0, column = 1)
		Button(palette, bitmap = "@./Subfiles/hflip.xbm", 
			command = self.hflip).grid(row = 1, column = 0)
		Button(palette, bitmap = "@./Subfiles/vflip.xbm", 
			command = self.vflip).grid(row = 1, column = 1)


	def bind(self, *args, **kwargs):	# used really often
		self.master.bind(*args, **kwargs)


	def curdisp(self, event = 0):	# displays cursor coords in frame
		item = self.edit_field.find_closest(event.x, event.y)[0]
		if 1 < item <= (1 + self.width*self.height):
			self.coords.set(str(divmod(item-2, self.width)))

	def show(self):	# updates the image on the side
		self.cur_pic.configure(data = self.cur_tile.tobitmap(3))

	def update(self):	# updates pixel canvas
		for row in range(self.height):
			for col in range(self.width):
				self.edit_field.itemconfigure(row*self.width + col + 2, 
					fill = colors[self.cur_tile[row, col]])
		self.show()


	def _set_states(self):
		self.ur_redos = (len(self.ur_stack) != self.ur_ptr + 1)
		undo = C_STATES[bool(self.ur_ptr)]
		redo = C_STATES[self.ur_redos]

		self.editmenu.entryconfig(0, state = undo)
		self.editmenu.entryconfig(1, state = redo)


